# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - PGP authentication using incoming MIME request bodies

    @copyright: 2001-2003 Juergen Hermann <jh@web.de>
                2003-2006 MoinMoin:ThomasWaldmann
                2007 MoinMoin:JohannesBerg
                2013 Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL, see COPYING for details.
"""

from MoinMoin import log
logging = log.getLogger(__name__)

from MoinMoin.user import User
from MoinMoin.auth import BaseAuth
from MoinSupport import getHeader
from MoinMessage import GPG, is_signed, is_encrypted, \
                        MoinMessageDecodingError, MoinMessageError
from MoinMessageSupport import get_homedir, get_username_for_fingerprint
from email.parser import Parser

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

class PGPAuth(BaseAuth):

    """
    Authenticate a user by inspecting the signature on a signed MIME message
    sent in the body of a request.
    """

    name = "pgp"

    def __init__(self, autocreate=False):

        "Initialise the authenticator using the given 'autocreate' setting."

        self.autocreate = autocreate
        BaseAuth.__init__(self)

    def request(self, request, user_obj, **kw):

        """
        Evaluate the 'request' given an existing 'user_obj', returning a tuple
        of the form (user, continue) where the user is either the supplied
        'user_obj' or a new user object, and where an indication of whether to
        continue the authentication process is also given.
        """

        _ = request.getText
        user = None

        # Always revalidate the identity if the provided user claims to be
        # identified using this method.

        if user_obj and user_obj.auth_method == self.name:
            user_obj = None

        # If something else authenticated before us, accept this.

        if user_obj:
            return user_obj, True

        logging.debug("request: %r" % request)

        # Read the request body and perform signature validation.

        # Need to obtain the message text and yet still make it available
        # normally in the request.

        content_length = getHeader(request, "Content-Length", "HTTP")
        if content_length:
            content_length = int(content_length)
        message_body = message_text = request.read(content_length)

        # Obtain a message from the text.

        message = Parser().parse(StringIO(message_text))

        try:
            homedir = get_homedir(request)
            gpg = GPG(homedir)

            try:
                # Encrypted messages must be decrypted first.

                if is_encrypted(message):
                    message_text = gpg.decryptMessage(message)
                    message = Parser().parse(StringIO(message_text))

                # Signed messages can be handled directly.

                if not is_signed(message):
                    return user_obj, True

                fingerprint, identity, content = gpg.verifyMessage(message)

            except MoinMessageDecodingError:
                logging.info("Incoming message was improperly encoded.")
                return user_obj, True

            except MoinMessageError:
                logging.warning("Incoming message was not verifiable.")
                return user_obj, True

        # Restore the request body using an extension to the request and
        # replaced methods.

        finally:
            extension = RequestExtension(request, StringIO(message_body))
            request.read = extension.read
            request._setup_args_from_cgi_form = extension._setup_args_from_cgi_form

        # Evaluate the result of the verification process.

        if fingerprint:
            username = get_username_for_fingerprint(request, fingerprint)

            # With a known username, temporarily switch user in order to make
            # the edit.

            if username:
                user = User(request, auth_method="pgp", auth_username=username)
                logging.debug("username: %r" % username)

        logging.debug("user: %r" % user)

        # Handle user autocreation.

        if user and self.autocreate:
            logging.debug("autocreating user")
            user.create_or_update()

        # Either return the identified user, if valid.

        if user and user.valid:
            logging.debug("returning valid user %r" % user)
            return user, True # True to get other methods called, too

        # Or return the supplied user object.

        else:
            logging.debug("returning %r" % user_obj)
            return user_obj, True

# Replacement request methods (based on the request_cgi.Request class).

class RequestExtension:

    "An extension to the request providing replacement methods."

    def __init__(self, request, body):
        self.request = request
        self.body = body

    def _setup_args_from_cgi_form(self):
        """ Override to create cgi form """
        form = cgi.FieldStorage(fp=self.body, environ=self.request.env, keep_blank_values=1)
        return RequestBase._setup_args_from_cgi_form(self.request, form)

    def read(self, n):
        """ Read from input stream. """
        if n is None:
            logging.warning("calling request.read(None) might block")
            return self.body.read()
        else:
            return self.body.read(n)

# vim: tabstop=4 expandtab shiftwidth=4
